#!/usr/bin/perl -w
#
#  Search utilities
#  A collection of utility functions to perform linear and binary searches
#
#  Copyright (c) 2001  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package Search;

#
#  Numeric sort
#
sub Nsort
{
   return sort {$a <=> $b} @_;
}

#
#  Alphanumeric sort
#
sub Asort
{
   return sort {$a cmp $b} @_;
}

#
#  Alphanumeric sort - no case
#
sub Usort
{
   return sort {uc($a) cmp uc($b)} @_;
}

#
#  Search for value in ordered list
#  Everything must be numeric
#  The array must be sorted
#  Example:  $i = binsearch($x,\@X);
#  Returns undef if not found
#  Returns $i where $X[$i] <= $x <= $X[$i+1]
#
sub binsearch
{
   (@_==2) || die "Wrong number of arguments for binsearch\n";
   my ($x,$X) = @_;

   my $n = scalar(@$X)-1;
   ($n>=0) || return;

   #  Forward sorted list
   if ($X->[$n] > $X->[0])
   {
      ($X->[0] <= $x && $x <= $X->[$n]) || return;
      my $i = 0;
      my $j = $n;
      while ($j-$i>1)
      {
         my $k = int(($i+$j)/2);
         if ($X->[$k] > $x)
         {
            $j = $k;
         }
         else
         {
            $i = $k;
         }
      }
      return $i;
   }
   #  Backward sorted list
   else
   {
      ($X->[0] >= $x && $x >= $X->[$n]) || return;
      my $i = 0;
      my $j = $n;
      while ($j-$i>1)
      {
         my $k = int(($i+$j)/2);
         if ($X->[$k] < $x)
         {
            $j = $k;
         }
         else
         {
            $i = $k;
         }
      }
   return $i;
   }
} 

#
#  Search for value in ordered list
#  Searches data numerically unless asked to do string comparisons
#  The array must be sorted
#  Returns undef if not found
#  Returns $i where $x eq/== $X[$i] if found
#
sub binfind
{
   (@_==3) || die "Wrong number of arguments for binfind\n";
   my ($x,$X,$type) = @_;

   my $n = scalar(@$X)-1;
   ($n>=0) || return;

   #  Use string comparisons
   if ($type eq "A")
   {
      ($X->[0] le $x && $x le $X->[$n]) || return;
      my $i = 0;
      my $j = $n;
      while ($j-$i>1)
      {
         my $k = int(($i+$j)/2);
         if ($X->[$k] eq $x)
         {
            $i = $j = $k;
         }
         elsif ($X->[$k] ge $x)
         {
            $j = $k;
         }
         else
         {
            $i = $k;
         }
      }
      if ($X->[$i] eq $x)
      {
         return $i;
      }
      elsif ($X->[$j] eq $x)
      {
         return $j;
      }
      else
      {
         return;
      }
   }
   #  Use case independent string comparisons
   elsif ($type eq "U")
   {
      $x = uc($x);
      (uc($X->[0]) le $x && $x le uc($X->[$n])) || return;
      my $i = 0;
      my $j = $n;
      while ($j-$i>1)
      {
         my $k = int(($i+$j)/2);
         if (uc($X->[$k]) eq $x)
         {
            $i = $j = $k;
         }
         elsif (uc($X->[$k]) ge $x)
         {
            $j = $k;
         }
         else
         {
            $i = $k;
         }
      }
      if (uc($X->[$i]) eq $x)
      {
         return $i;
      }
      elsif (uc($X->[$j]) eq $x)
      {
         return $j;
      }
      else
      {
         return;
      }
   }
   #  Use numeric comparisons
   elsif ($type eq "N")
   {
      ($X->[0] <= $x && $x <= $X->[$n]) || return;
      my $i = 0;
      my $j = $n;
      while ($j-$i>1)
      {
         my $k = int(($i+$j)/2);
         if ($X->[$k] == $x)
         {
            $i = $j = $k;
         }
         elsif ($X->[$k] >= $x)
         {
            $j = $k;
         }
         else
         {
            $i = $k;
         }
      }
      if ($X->[$i] == $x)
      {
         return $i;
      }
      elsif ($X->[$j] == $x)
      {
         return $j;
      }
      else
      {
         return;
      }
   }
   else
   {
      die "Unknown binfind type $type\n";
   }
} 

#
#  Search for value in unordered list
#  Searches data numerically unless asked to do string comparisons
#  Returns undef if not found
#  Returns $i where $x eq/== $X[$i] if found
#
sub find
{
   my ($x,$X,$type) = @_;

   my $n = scalar(@$X);
   ($n>0) || return;

   #  Use string comparisons
   if ($type eq "A")
   {
      for (my $i=0;$i<$n;$i++)
      {
         ($X->[$i] ne $x) || return $i;
      }
      return;
   }
   #  Use case independent string comparisons
   elsif ($type eq "U")
   {
      $x = uc($x);
      for (my $i=0;$i<$n;$i++)
      {
         (uc($X->[$i]) ne $x) || return $i;
      }
      return;
   }
   #  Use numeric comparisons
   elsif ($type eq "N")
   {
      for (my $i=0;$i<$n;$i++)
      {
         ($X->[$i] != $x) || return $i;
      }
      return;
   }
   else
   {
      die "Unknown find type $type\n";
   }
} 

1;
